# This file is part of Online Services Jenkaas.
#
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals


import argparse
import errno
import logging
import os
import sys


from olsconfig import registries
from olsvms import (
    errors,
    subprocesses,
)


import olsjenkaas
from olsjenkaas import (
    config,
    containers,
    launchpad,
)


logger = logging.getLogger(__name__)


def setup_logging(level=None):
    if level is None:
        level = logging.ERROR
    if level == logging.DEBUG:
        log_format = '%(asctime)s %(name)s %(levelname)s %(message)s'
    else:
        log_format = '%(levelname)s %(message)s'
    logging.basicConfig(level=level, format=log_format)


def ensure_dir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        # python2 doesn't provide the exist_ok=True to makedirs
        if e.errno != errno.EEXIST:
            raise


class ArgParser(argparse.ArgumentParser):
    """An argument parser for the ols-jenkaas script."""

    def __init__(self, name, description):
        script_name = 'ols-jenkaas'
        super(ArgParser, self).__init__(
            prog='{} {}'.format(script_name, name),
            description=description)

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        :params args: Defaults to sys.argv[1:].

        :param out: Defaults to sys.stdout.

        :param err: Defaults to sys.stderr.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


class CommandRegistry(registries.Registry):
    """A registry specialized for commands."""

    def register(self, cmd):
        super(CommandRegistry, self).register(
            cmd.name, cmd, help_string=cmd.description)


# All commands are registered here, defining what run() supports
command_registry = CommandRegistry()


class Command(object):

    name = 'Must be set by daughter classes'
    description = 'Must be set by daughter classes'

    def __init__(self, out=None, err=None):
        """Command constructor.

        :param out: A stream for command output.

        :param err: A stream for command errors.
        """
        if out is None:
            out = sys.stdout
        if err is None:
            err = sys.stderr
        self.out = out
        self.err = err
        self.parser = ArgParser(self.name, self.description)

    def parse_args(self, args):
        self.options = self.parser.parse_args(args, self.out, self.err)
        return self.options


class Help(Command):

    name = 'help'
    description = 'Describe ols-jenkaas commands.'

    def __init__(self, **kwargs):
        super(Help, self).__init__(**kwargs)
        self.parser.add_argument(
            'commands', metavar='COMMAND', nargs='*',
            help='Display help for each command'
            ' (All if none is given).')

    def run(self):
        if not self.options.commands:
            self.out.write('Available commands:\n')
            for cmd_name in command_registry.keys():
                cmd_class = command_registry.get(cmd_name)
                self.out.write('\t{}: {}\n'.format(cmd_class.name,
                                                   cmd_class.description))
            return
        for cmd_name in self.options.commands:
            try:
                cmd_class = command_registry.get(cmd_name)
            except KeyError:
                cmd_class = None
            if cmd_class:
                cmd = cmd_class()
                cmd.parser.print_help(self.out)
            else:
                msg = '"{}" is not a known command\n'.format(cmd_name)
                self.err.write(msg)
        return 0


command_registry.register(Help)


class ApprovedBzrProposals(Command):

    name = 'approved-bzr-proposals'
    description = 'Check for approved bzr merge proposals.'

    def __init__(self, **kwargs):
        super(ApprovedBzrProposals, self).__init__(**kwargs)
        self.parser.add_argument(
            'target', metavar='TARGET',
            help='The bazaar target branch for the merge proposals.')

    def run(self):
        self.lp = launchpad.Launchpad(config.VmStack(None))
        target_branch = self.lp.get_bzr_branch(self.options.target)
        if target_branch is None:
            logger.error('{} is not a bzr branch'.format(
                self.options.target))
            return 1
        logger.info('Searching proposals for {}'.format(
            target_branch.web_link))
        for proposal in target_branch.landing_candidates:
            logger.info('Looking at {}'.format(proposal.web_link))
            unique_name = proposal.target_branch.unique_name
            if unique_name != target_branch.unique_name:
                logger.info('Ignoring, target is {}'.format(
                    unique_name))
                continue
            queue_status = proposal.queue_status
            if queue_status != 'Approved':
                logger.info('Ignoring, status is {}'.format(queue_status))
                continue
            # The proposal is approved
            logger.info('{} is approved'.format(proposal.web_link))
            return 0
        # None of the proposals are approved
        logger.error('No proposal is approved')
        return 1


command_registry.register(ApprovedBzrProposals)


class ApprovedGitProposals(Command):

    name = 'approved-git-proposals'
    description = 'Check for approved git merge proposals.'

    def __init__(self, **kwargs):
        super(ApprovedGitProposals, self).__init__(**kwargs)
        # FIXME: Shouldn't we mention the branch inside the repository there ?
        # Even if defaulting to 'master' -- vila 2016-12-07
        self.parser.add_argument(
            'target', metavar='TARGET',
            help='The target git repository for the merge proposals.')

    def run(self):
        self.lp = launchpad.Launchpad(config.VmStack(None))
        target_git_repo = self.lp.get_git_repo(self.options.target)
        if target_git_repo is None:
            logger.error('{} is not a git repository'.format(
                self.options.target))
            return 1
        for proposal in target_git_repo.landing_candidates:
            logger.info('Looking at {}'.format(proposal.web_link))
            unique_name = proposal.target_git_repository.unique_name
            if unique_name != target_git_repo.unique_name:
                logger.info('Ignoring, target is {}'.format(
                    unique_name))
                continue
            queue_status = proposal.queue_status
            if queue_status != 'Approved':
                logger.info('Ignoring, status is {}'.format(queue_status))
                continue
            # The proposal is approved
            logger.info('{} is approved'.format(proposal.web_link))
            return 0
        # None of the Proposals are approved
        logger.error('No proposal is approved')
        return 1


command_registry.register(ApprovedGitProposals)


class CommandWithContainer(Command):

    def __init__(self, **kwargs):
        super(CommandWithContainer, self).__init__(**kwargs)
        # The container where commands are run
        self.worker = None

    def get_mounts_and_remote(self, local_dir):
        """Build mount points and remote working directory.

        :param local_dir: The local directory needed inside the worker.

        :return: A (mounts, remote) tuple where 'mounts' is a list of mount
            points and 'remote' the absolute path for the directory inside the
            container.
        """
        mounts = ['~/workspace:/workspace']
        rel_path = os.path.relpath(
            local_dir,
            os.path.realpath(os.path.expanduser('~/workspace')))
        # FIXME: Arguably a more restricted view should be exposed but at least
        # the slave home directory is not exposed in the container.
        # -- vila 2017-02-07
        remote_work_dir = os.path.join('/workspace', rel_path)
        return mounts, remote_work_dir

    def create_worker(self, backing_container_name, name=None,
                      mounts=None):
        """Create and start a worker.

        :param backing_container_name: The base container defined in
            ols-vms.conf.

        :param name: The worker name (defaults to worker-<base name>-<pid> if
            not specified').

        :params mounts: An optional list of mount '<local>:<remote>' pairs.

        """
        if name is None:
            name = 'worker-{}-{}'.format(backing_container_name, os.getpid())
        worker = containers.Worker(name, backing_container_name,
                                   out=self.out, err=self.err)
        ret = worker.start(mounts)
        # Only remember the worker if it can start (for cleanup), otherwise, an
        # error is displayed.
        self.worker = worker
        return ret

    def delete_worker(self):
        self.worker.stop()

    def run_in_container(self, command, with_creds=False):
        return self.worker.raw_shell(command, with_agent=with_creds)


class RunInWorker(CommandWithContainer):

    name = 'run-in-worker'
    description = 'Run commands in a worker.'

    def __init__(self, **kwargs):
        super(RunInWorker, self).__init__(**kwargs)
        self.parser.add_argument(
            'container_name', metavar='CONTAINER',
            help='The base container to build the worker.')
        self.parser.add_argument(
            'commands', metavar='COMMAND', nargs='+',
            help='A command to run in the worker.'
            ' Prefix with with-credentials to allow authenticated access.')

    def run(self):
        ret = 1  # In case worker creation fails
        try:
            local_work_dir = os.getcwd()
            mounts, remote_work_dir = self.get_mounts_and_remote(
                local_work_dir)
            logger.info('Creating worker...')
            self.create_worker(self.options.container_name, mounts=mounts)
            logger.info('{} setup on top of {}'.format(
                self.worker.name, self.worker.backing_name))
            ret = 0
            creds_prefix = 'with-credentials '
            # FIXME: It would be nice to get the commands from the config,
            # jenkaas.commands ? {project}.commands ? {job}.commands ?
            # -- vila 2017-02-08
            for command in self.options.commands:
                if command.startswith(creds_prefix):
                    with_creds = True
                    command = command[len(creds_prefix):]
                else:
                    with_creds = False
                msg = 'Running `{}` in a {} worker'.format(
                    command, self.options.container_name)
                if with_creds:
                    msg += ' with credentials'
                logger.info(msg)
                ret = self.run_in_container(
                    ['cd', remote_work_dir, '&&', command],
                    with_creds)
                if ret:
                    logger.error('{} failed in {}'.format(command,
                                                          self.worker.name))
                    return ret
            return ret
        finally:
            if self.worker is not None:
                self.delete_worker()


command_registry.register(RunInWorker)


# FIXME: Add support for pre-requisites -- vila 2017-01-30
class LandApprovedProposal(CommandWithContainer):

    def __init__(self, **kwargs):
        super(LandApprovedProposal, self).__init__(**kwargs)
        self.parser.add_argument(
            'target', metavar='TARGET',
            help=self.target_help)
        self.parser.add_argument(
            'work_dir', metavar='WORKINGDIR',
            help='Where the target is checked out.')
        self.parser.add_argument(
            'container_name', metavar='CONTAINER',
            help='The container name where gating tests are run.')
        self.parser.add_argument(
            'setup_cmd', metavar='SETUP',
            help='The command to run in the container to setup the project.')
        self.parser.add_argument(
            'test_cmd', metavar='TEST',
            help='The command to run the tests in the container.')

    def vcs(self, vcs_cmd_name, args):
        try:
            vcs_cmd = [vcs_cmd_name] + args
            return subprocesses.run(vcs_cmd)
        except errors.CommandError:
            logger.exception('{} failed'.format(' '.join(vcs_cmd)))
            raise

    # FIXME: Every method with 'proposal' in its name or as a paremeter is
    # yelling: "composition !" -- vila 2017-02-06
    def get_target(self, url):
        """Get the launchpad object the proposals are targeted.

        :param url: Describing the target
        """
        raise NotImplementedError(self.get_target)

    def proposal_comment_subject(self, proposal):
        """Build a comment subject for the given proposal.

        :param proposal: The merge proposal needing a comment.
        """
        raise NotImplementedError(self.proposal_comment_subject)

    def setup_merge_target(self, work_dir, target_url):
        """Setup a branch in a specified directory.

        :param work_dir: Path to install the branch at.

        :param target_url: Branch url.
        """
        # Inherited by daughter classes, cleanup first.
        if os.path.exists(work_dir):
            subprocesses.run(['rm', '-fr', work_dir])

    def merge_proposed_branch(self, work_dir, proposal):
        """Merge the proposed branch into the target.

        :param work_dir: Path where the merge should occur.

        :param proposal: Launchpad proposal to merge.
        """
        raise NotImplementedError(self.merge_proposed_branch)

    def commit_proposal(self, work_dir, proposal):
        """Commit the current working tree.

        :param work_dir: Path where the commit should occur.

        :param proposal: Launchpad proposal already merged.
        """
        raise NotImplementedError(self.commit_proposal)

    def push_merged_proposal(self, work_dir, target_url):
        """Push the merged proposal to its target.

        :param work_dir: Path from where the push is done.

        :param target_url: Branch url to push to.
        """
        raise NotImplementedError(self.push_merged_proposal)

    def reject_proposal(self, proposal, reason):
        subject = self.proposal_comment_subject(proposal)
        # Include a reference to the job url if running under jenkins
        content = reason
        build_url = os.environ.get('BUILD_URL', None)
        if build_url:
            content += '\n' + build_url
        proposal.createComment(subject=subject, content=content)
        proposal.setStatus(status='Needs review',
                           revid=proposal.reviewed_revid)

    def land_proposal(self, proposal, target_url):

        class Rejected(Exception):

            def __init__(self, reason):
                self.reason = reason

        logger.info('Trying to land {}'.format(proposal.web_link))
        try:
            if not proposal.commit_message:
                raise Rejected('A commit message must be set')
            if not launchpad.is_approved(proposal, self.target):
                raise Rejected('Voting criteria not met')
            local_work_dir = os.path.realpath(self.options.work_dir)
            logger.info('Getting {}'.format(target_url))
            self.setup_merge_target(local_work_dir, target_url)
            logger.info('Merging {}'.format(proposal.web_link))
            try:
                merged = self.merge_proposed_branch(local_work_dir, proposal)
            except errors.CommandError:
                raise Rejected('Merging failed')
            if not merged:
                logger.error('Nothing to merge from {}'.format(
                    proposal.web_link))
                raise Rejected('There was nothing to merge')
            mounts, remote_work_dir = self.get_mounts_and_remote(
                local_work_dir)
            logger.info('Creating worker...')
            with olsjenkaas.working_directory(local_work_dir):
                self.create_worker(self.options.container_name, mounts=mounts)
            logger.info('{} setup on top of {}'.format(
                self.worker.name, self.worker.backing_name))
            logger.info('Running {}'.format(self.options.setup_cmd))
            if self.run_in_container(
                    ['cd', remote_work_dir, '&&', self.options.setup_cmd],
                    with_creds=True):
                logger.error('{} failed'.format(self.options.setup_cmd))
                raise Rejected('Project setup failed')
            logger.info('Running {}'.format(self.options.test_cmd))
            if self.run_in_container(
                    ['cd', remote_work_dir, '&&', self.options.test_cmd]):
                logger.error('{} failed'.format(self.options.test_cmd))
                raise Rejected('Running landing tests failed')
            self.commit_proposal(local_work_dir, proposal)
            self.push_merged_proposal(local_work_dir, target_url)
            # FIXME: Once webhooks are used, lp can be trusted to do that (with
            # some delay), revert this once this has been verified
            # -- vila 2017-03-30
            proposal.setStatus(status='Merged')
            logger.info('{} has landed'.format(proposal.web_link))
            return 0
        except Rejected as e:
            logger.info('{} could not be landed: {}'.format(
                proposal.web_link, e.reason))
            self.reject_proposal(proposal, e.reason)
            return 1
        finally:
            if self.worker is not None:
                self.delete_worker()

    def run(self):
        self.lp = launchpad.Launchpad(config.VmStack(None))
        self.target = self.get_target(self.options.target)
        if self.target is None:
            logger.error(self.wrong_target.format(self.options.target))
            return 1
        logger.info('Searching proposals for {}'.format(
            self.target.web_link))
        for proposal in self.target.landing_candidates:
            logger.debug('Looking at {}'.format(proposal.web_link))
            unique_name = self.proposal_unique_name(proposal)
            if unique_name != self.target.unique_name:
                logger.debug('Ignoring, target is {}'.format(
                    unique_name))
                continue
            queue_status = proposal.queue_status
            if queue_status != 'Approved':
                logger.debug('Ignoring, status is {}'.format(queue_status))
                continue
            # The proposal is approved
            logger.info('{} is approved'.format(proposal.web_link))
            return self.land_proposal(proposal, self.options.target)
        # None of the proposals are approved. Returning an error here cause job
        # failures for little added value.
        logger.error('No proposal is approved')
        return 0


class LandApprovedBzrProposal(LandApprovedProposal):

    name = 'land-approved-bzr-proposal'
    description = 'Land an approved bzr merge proposal.'
    target_help = 'The bazaar target branch for the merge proposals.'
    wrong_target = '{} is not a bzr branch'

    def __init__(self, **kwargs):
        super(LandApprovedBzrProposal, self).__init__(**kwargs)

    def bzr(self, args):
        return self.vcs('bzr', args)

    def get_target(self, url):
        return self.lp.get_bzr_branch(url)

    def proposal_comment_subject(self, proposal):
        # Uses display_name (lp:xxx)
        return 'Re: [Merge] {} into {}'.format(
            proposal.source_branch.display_name,
            proposal.target_branch.display_name)

    def setup_merge_target(self, work_dir, target_url):
        super(LandApprovedBzrProposal, self).setup_merge_target(
            work_dir, target_url)
        if not os.path.exists(work_dir):
            # The parent directory must exists
            ensure_dir(os.path.dirname(work_dir))
        return self.bzr(['branch', target_url, work_dir])

    def merge_proposed_branch(self, work_dir, proposal):
        with olsjenkaas.working_directory(work_dir):
            lp_url = proposal.source_branch.bzr_identity
            # FIXME: This is where
            # https://bugs.launchpad.net/ols-jenkaas/+bug/1663246 needs more
            # work (and the git implementation too). A possible fix is just to
            # *not* specify '-r', proposal.reviewed_revid but that means
            # ignoring what has been top-approved and as such is worth
            # discussing whether it's the right thing to do. -- vila 2017-02-09
            ret, out, err = self.bzr(
                ['merge', lp_url, '-r', proposal.reviewed_revid])
            if err == 'Nothing to do.\n':
                # There was nothing to merge
                return False
            else:
                return True

    def get_author(self, work_dir, revid):
        # FIXME: Not explicitly tested -- vila 2017-01-26
        with olsjenkaas.working_directory(work_dir):
            log = self.bzr(['log', '-l1', '-r', revid])[1]
            for line in log.splitlines():
                if line.startswith('author:'):
                    return line[len('author: '):].rstrip('\n')
                if line.startswith('committer:'):
                    return line[len('committer: '):].rstrip('\n')
            return 'unknown'

    def commit_proposal(self, work_dir, proposal):
        author = self.get_author(work_dir, proposal.reviewed_revid)
        # Decorate commits (not as sexy as lp but good enough)
        commit_msg = (proposal.commit_message + '\n\n' +
                      'Merged from {}'.format(proposal.web_link))
        with olsjenkaas.working_directory(work_dir):
            self.bzr(['commit', '-m', commit_msg, '--author', author])

    def push_merged_proposal(self, work_dir, target_url):
        with olsjenkaas.working_directory(work_dir):
            self.bzr(['push', target_url])

    def proposal_unique_name(self, proposal):
        return proposal.target_branch.unique_name


command_registry.register(LandApprovedBzrProposal)


class LandApprovedGitProposal(LandApprovedProposal):

    name = 'land-approved-git-proposal'
    description = 'Land an approved git merge proposal.'
    target_help = 'The target git repository branch for the merge proposals.'
    wrong_target = '{} is not a git repository'

    def __init__(self, **kwargs):
        super(LandApprovedGitProposal, self).__init__(**kwargs)
        # FIXME: Shouldn't we mention the branch inside the repository there ?
        # Even if defaulting to 'master' -- vila 2016-12-07

    def git(self, args):
        return self.vcs('git', args)

    def get_target(self, url):
        return self.lp.get_git_repo(url)

    def proposal_comment_subject(self, proposal):
        # Uses unique_name (xxx)
        proposed_url = '{}:{}'.format(
            proposal.source_git_repository.unique_name,
            proposal.source_git_path[len('refs/heads/'):])
        target_url = '{}:{}'.format(
            proposal.target_git_repository.unique_name,
            proposal.target_git_path[len('refs/heads/'):])
        return 'Re: [Merge] {} into {}'.format(proposed_url, target_url)

    def setup_merge_target(self, work_dir, target_url, branch_name='master'):
        super(LandApprovedGitProposal, self).setup_merge_target(
            work_dir, target_url)
        self.git(['clone', target_url, work_dir])
        with olsjenkaas.working_directory(work_dir):
            self.git(['checkout', branch_name])

    def merge_proposed_branch(self, work_dir, proposal):
        with olsjenkaas.working_directory(work_dir):
            self.git(['fetch', proposal.source_git_repository.git_identity,
                      proposal.source_git_path])
            ret, out, err = self.git(['merge', '--no-ff', '--no-commit',
                                      proposal.reviewed_revid])
            if out == 'Already up-to-date.\n':
                # There was nothing to merge
                return False
            else:
                return True

    def get_author(self, work_dir, revid):
        with olsjenkaas.working_directory(work_dir):
            author = self.git(['log', '-1', '--format=%aN <%aE>',
                               revid])[1].rstrip('\n')
        return author

    def commit_proposal(self, work_dir, proposal):
        author = self.get_author(work_dir, proposal.reviewed_revid)
        # Decorate commits (not as sexy as lp but good enough)
        commit_msg = (proposal.commit_message + '\n\n' +
                      'Merged from {}'.format(proposal.web_link))
        with olsjenkaas.working_directory(work_dir):
            self.git(['commit', '-m', commit_msg, '--author', author])

    def push_merged_proposal(self, work_dir, target_url, branch='master'):
        with olsjenkaas.working_directory(work_dir):
            self.git(['push', target_url, branch])

    def proposal_unique_name(self, proposal):
        return proposal.target_git_repository.unique_name


command_registry.register(LandApprovedGitProposal)


def run(args=None, out=None, err=None):
    if args is None:
        args = sys.argv[1:]
    if not args:
        cmd = Help()
        cmd_name = 'help'
    else:
        cmd_name = args[0]
        args = args[1:]
        try:
            cmd_class = command_registry.get(cmd_name)
            cmd = cmd_class(out=out, err=err)
        except KeyError:
            cmd = Help(out=out, err=err)
            args = [cmd_name]
    try:
        cmd.parse_args(args)
        return cmd.run()
    except Exception as e:
        logger.debug('{} failed'.format(cmd_name), exc_info=True)
        logger.error('{} failed: {!r}'.format(cmd_name, e))
        return -1
