# This file is part of Online Services Jenkaas.
#
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import logging
import os
import time


from olstests import fixtures
from olsvms import subprocesses


from olsjenkaas import (
    config,
    launchpad,
)
from olsjenkaas.tests import features
from olsjenkaas.tests.branches import (
    bzr,
    git,
)


# Useful shorcuts
build_tree = fixtures.build_tree
patch = fixtures.patch


def isolate_from_disk(test):
    """Provide an isolated disk-based environment.

    A $HOME directory is setup so tests can setup arbitrary files including
    config ones.
    """
    fixtures.set_uniq_cwd(test)
    # Share the lxd certificate. We could generate one on the fly but assume
    # instead that one is already available since we require the lxc client
    # already.
    lxd_conf_dir = os.path.expanduser('~/.config/lxc')
    # Preserve some options from the outer environment
    outer = config.VmStack(None)
    setup_known_hosts_path = outer.get('jenkaas.setup_known_hosts')
    lp_login = outer.get('launchpad.login')
    fixtures.isolate_from_env(test)
    # Isolate tests from the user environment
    test.home_dir = os.path.join(test.uniq_dir, 'home')
    os.mkdir(test.home_dir)
    os.environ['HOME'] = test.home_dir
    os.environ['LXD_CONF'] = lxd_conf_dir
    # Inject preserved options
    inner = config.VmStack(None)
    inner.set('jenkaas.setup_known_hosts', setup_known_hosts_path)
    inner.set('launchpad.login', lp_login)
    # Override some options for test purposes (generally the defaults for all
    # jenkaas workers and as such needing some care for tests to not stomp on
    # each other feet). Also, all expected uses of containers start in a
    # working directory below ~/workspace so mimic that behavior.
    test.workspace = os.path.join(test.home_dir, 'workspace')
    os.mkdir(test.workspace)
    os.chdir(test.workspace)
    # inner.set('jenkaas.worker.mounts', test.workspace + ':/workspace')
    inner.store.save()
    inner.store.unload()


def use_qastaging(test):
    features.test_requires(test, features.qastaging_token)
    # Get original path before isolation
    credentials_path = features.qastaging_token.path
    isolate_from_disk(test)
    # Setup to use qastaging launchpad API and friends
    conf = config.VmStack(None)
    conf.set('launchpad.service_root', 'qastaging')
    conf.set('launchpad.credentials', '~/.config/ols-vms/launchpad-qastaging')
    conf.store.save()
    conf.store.unload()
    with open(credentials_path) as fin:
        with open(conf.get('launchpad.credentials'), 'w') as fout:
            fout.write(fin.read())
    return launchpad.Launchpad(config.VmStack(None))


def override_logging(test, stream_out, level=logging.INFO, fmt='%(message)s'):
    """Setup a logging handler, restoring the actual handlers after the test.

    This assumes a logging setup where handlers are added to the root logger
    only.

    :param stream_out: A stream where log messges will be recorded.

    :param level: The logging level to be active during the test.

    :param fmt: The logging format to use during the test.

    """
    root_logger = logging.getLogger(None)
    # Using reversed() below ensures we can modify root_logger.handlers as well
    # as providing the handlers in the right order for cleanups.
    for handler in reversed(root_logger.handlers):
        root_logger.removeHandler(handler)
        test.addCleanup(root_logger.addHandler, handler)
    # Install the new handler
    new_handler = logging.StreamHandler(stream_out)
    test.addCleanup(root_logger.removeHandler, new_handler)
    root_logger.addHandler(new_handler)
    # Install the new level, restoring the actual one after the test
    test.addCleanup(root_logger.setLevel, root_logger.level)
    root_logger.setLevel(level)


def set_bzr_identity(test):
    """Set the bzr whoami."""
    features.test_requires(test, features.bzr_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    subprocesses.run(['bzr', 'whoami', features.bzr_identity.identity])


def set_bzr_launchpad_access(test):
    """Set the bzr launchpad login."""
    features.test_requires(test, features.launchpad_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    # All tests run against qastaging
    plugin_dir = os.path.join(test.uniq_dir, 'home', '.bazaar', 'plugins')
    os.makedirs(plugin_dir)
    with open(os.path.join(plugin_dir, 'qastagingonly.py'), 'w') as f:
        f.write('from bzrlib.plugins.launchpad import lp_directory\n')
        f.write('lp_directory.LaunchpadService.DEFAULT_INSTANCE = "qastaging"')
    subprocesses.run(['bzr', 'launchpad-login',
                      features.launchpad_identity.identity])
    # bazaar.qastaging.launchpad.net must be in ~/.ssh/known_hosts


def set_git_identity(test):
    """Set the git user name and email."""
    features.test_requires(test, features.git_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    subprocesses.run(['git', 'config', '--global', 'user.name',
                      features.git_identity.name])
    subprocesses.run(['git', 'config', '--global', 'user.email',
                      features.git_identity.email])


def set_git_launchpad_access(test):
    """Configure git to access launchpad."""
    features.test_requires(test, features.launchpad_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    # All tests run against qastaging
    # git.qastaging.paddev.net must be in ~/.ssh/known_hosts
    subprocesses.run(
        ['git', 'config', '--global',
         'url.ssh://{}@git.qastaging.paddev.net/.insteadOf'.format(
             features.launchpad_identity.identity),
         'lp:'])
    subprocesses.run(
        ['git', 'config', '--global', 'push.default', 'simple'])


class VcsBranchHandler(object):
    """Abstract base class to create local branches."""

    # Must be set by daughter classes
    factory = None
    """The class responsible for handling branches (or repositories)."""
    vcs_dir = None
    """A directory created by the vcs for storing its files."""

    def set_identity(self, test):
        """Setup the vcs identity."""
        raise NotImplementedError(self.set_identify)

    def set_launchpad_access(self, test):
        """Setup launchpad access for vcs."""
        raise NotImplementedError(self.set_launchpad_access)

    def lp_branch_url(self, stem):
        """Create a full launchpad url from a stem.

        This should embded the process pid so tests can be run concurrently
        without collision. A single (dedicated but arbitrary) project
        (ols-jenkaas-test-project) should be used to keep things isolated but
        simple. The MP url is unique as long as a single user doesn't attempt
        to run the test suite from different hosts at the same time and being
        unlucky enough to collide on the process id.

        """
        raise NotImplementedError(self.lp_test_url)

    def get_branch_from_lp(self, lp, url):
        """Get the launchpad object associated with the branch."""
        raise NotImplementedError(self.set_identify)


class BzrBranchHandler(VcsBranchHandler):

    factory = bzr.Branch
    vcs_dir = '.bzr'

    def set_identity(self, test):
        set_bzr_identity(test)

    def set_launchpad_access(self, test):
        set_bzr_launchpad_access(test)

    def lp_branch_url(self, stem):
        return 'lp:~{}/ols-jenkaas-test-project/{}-{}'.format(
            features.launchpad_identity.identity,
            stem,
            os.getpid())

    def get_branch_from_lp(self, lp, url):
        return lp.get_bzr_branch(url)


class GitRepositoryHandler(VcsBranchHandler):

    factory = git.Repository
    vcs_dir = '.git'

    def set_identity(self, test):
        set_git_identity(test)

    def set_launchpad_access(self, test):
        set_git_launchpad_access(test)

    def lp_branch_url(self, stem):
        return 'lp:~{}/ols-jenkaas-test-project/+git/{}-{}'.format(
            features.launchpad_identity.identity,
            stem,
            os.getpid())

    def get_branch_from_lp(self, lp, url):
        return lp.get_git_repo(url)


class VcsProposalHandler(object):
    """Abstract base class to interact with proposals."""

    def create(self, test, source, target, message):
        """Create a merge proposal of source into target."""
        raise NotImplementedError(self.create_proposal)

    def build_comment_subject(self, proposal):
        """Build a standard subject for proposal comment."""
        raise NotImplementedError(self.build_comment_subject)


class BzrProposalHandler(VcsProposalHandler):

    def create(self, test, proposed, target, message):
        return proposed.createMergeProposal(
            target_branch=target,
            initial_comment=message,
            commit_message=message,
            needs_review=True)

    # FIXME: duplicated in LandApprovedBzrProposal.proposal_comment_subject
    # -- vila 2017-01-25
    def build_comment_subject(self, proposal):
        # Uses display_name (lp:xxx)
        return 'Re: [Merge] {} into {}'.format(
            proposal.source_branch.display_name,
            proposal.target_branch.display_name)


class GitProposalHandler(VcsProposalHandler):

    def get_ref(self, repo, branch):
        ref = repo.getRefByPath(path=branch)
        while ref is None:
            # FIXME: This doesn't seem to happen frequently but if it does, we
            # should switch to some expotential backoff (ols-vms has a timeouts
            # module for that) -- vila 2017-01-26
            time.sleep(2)
            ref = repo.getRefByPath(path='master')
        return ref

    def create(self, test, proposed, target, message):
        proposed_git_ref = self.get_ref(proposed, 'master')
        test.assertEqual('refs/heads/master', proposed_git_ref.path)
        target_git_ref = self.get_ref(target, 'master')
        test.assertEqual('refs/heads/master', target_git_ref.path)
        return proposed_git_ref.createMergeProposal(
            merge_target=target_git_ref,
            initial_comment=message,
            commit_message=message,
            needs_review=True)

    # FIXME: duplicated in LandApprovedGitProposal.proposal_comment_subject
    # -- vila 2017-01-25
    def build_comment_subject(self, proposal):
        # Uses unique_name (xxx)
        proposed_url = '{}:{}'.format(
            proposal.source_git_repository.unique_name,
            proposal.source_git_path[len('refs/heads/'):])
        target_url = '{}:{}'.format(
            proposal.target_git_repository.unique_name,
            proposal.target_git_path[len('refs/heads/'):])
        return 'Re: [Merge] {} into {}'.format(
            proposed_url,
            target_url)


def setup_proposal_target(test):
    """Create a local target branch to build proposals."""
    branch = test.branch_handler.factory('target')
    test.addCleanup(branch.delete)
    branch.create()
    branch.update('Something', 'file: foo\nfoo content\n')
    test.target_url = test.branch_handler.lp_branch_url('target')
    branch.push(test.target_url)
    test.target_lp_branch = test.branch_handler.get_branch_from_lp(
        test.lp, test.target_url)
    test.assertIsNot(None, test.target_lp_branch)
    test.addCleanup(test.target_lp_branch.lp_delete)


def create_proposal(test, name, approved=False, commit_msg=None):
    """Create a merge proposal from a branch.

    :param test: A test case with a 'proposal_handler', 'branch_handler',
        'target_url' and 'target_lp_branch' attributes. A 'proposed_lp_branch'
        attribute will be created. See setup_proposal_target() to create
        'target_url' and 'target_lp_branch'.

    :param name: The branch name.

    :param approved: Should the proposal be approved.

    :param commit_msg: Commit message for the proposal. A default one is
        created is None is provided.

    """
    proposed = test.branch_handler.factory(name)
    # Strictly speaking we should create from the lp branch but we just
    # pushed 'target' there so this avoids a useless roundtrip.
    proposed.create_from('target')
    vcs_commit_msg = '{} purpose'.format(name)
    if commit_msg is None:
        commit_msg = vcs_commit_msg
    proposed.update(vcs_commit_msg,
                    'file: {}\n{} content\n'.format(name, name))
    proposed_lp_url = test.branch_handler.lp_branch_url(name)
    proposed.push(proposed_lp_url)
    test.proposed_lp_branch = test.branch_handler.get_branch_from_lp(
        test.lp, proposed_lp_url)
    test.addCleanup(test.proposed_lp_branch.lp_delete)
    proposal = test.proposal_handler.create(
        test, test.proposed_lp_branch, test.target_lp_branch, commit_msg)
    if approved:
        subject = test.proposal_handler.build_comment_subject(proposal)
        proposal.createComment(subject=subject, vote='Approve',
                               content='Well done!')
        proposed_rev_id = proposed.tip_commit_id()
        proposal.setStatus(status='Approved', revid=proposed_rev_id)
    return proposal


def define_uniq_container(test):
    """To isolate tests from each other, created containers need a unique name.

    To keep those names legal and still user-readable we use the class name and
    the test method name. The process id is added so that the same test suite
    can be run on the same host in several processes.
    """
    container_name = '{}-{}-{}'.format(test.__class__.__name__,
                                       test._testMethodName,
                                       os.getpid())
    # '_' are not valid in hostnames
    test.container_name = container_name.replace('_', '-')
    conf = config.VmStack(test.container_name)
    conf.set('vm.class', 'lxd')
    conf.set('vm.update', 'False')
    conf.set('vm.release', 'xenial')
    conf.set('vm.architecture', 'amd64')
    conf.store.save()
    conf.store.unload()
