# This file is part of Online Services Jenkaas.
#
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals


import io
import os
import unittest


from olstests import (
    assertions,
    scenarii,
)
from olsvms import config


from olsjenkaas import (
    commands,
    containers,
)
from olsjenkaas.tests import fixtures


load_tests = scenarii.load_tests_with_scenarios


class TestHelpOptions(unittest.TestCase):

    def setUp(self):
        super(TestHelpOptions, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        help_cmd = commands.Help(out=self.out, err=self.err)
        return help_cmd.parse_args(args)

    def test_defaults(self):
        ns = self.parse_args([])
        self.assertEqual([], ns.commands)

    def test_single_command(self):
        ns = self.parse_args(['setup'])
        self.assertEqual(['setup'], ns.commands)

    def test_several_commands(self):
        ns = self.parse_args(['help', 'setup', 'not-a-command'])
        self.assertEqual(['help', 'setup', 'not-a-command'], ns.commands)


class TestHelp(unittest.TestCase):

    def setUp(self):
        super(TestHelp, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()

    def assertHelp(self, expected, args=None):
        if args is None:
            args = []
        help_cmd = commands.Help(out=self.out, err=self.err)
        help_cmd.parse_args(args)
        help_cmd.run()
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.out.getvalue())

    def test_help_alone(self):
        self.assertHelp('''\
Available commands:
\tapproved-bzr-proposals: Check for approved bzr merge proposals.
\tapproved-git-proposals: Check for approved git merge proposals.
\thelp: Describe ols-jenkaas commands.
\tland-approved-bzr-proposal: Land an approved bzr merge proposal.
\tland-approved-git-proposal: Land an approved git merge proposal.
\trun-in-worker: Run commands in a worker.
''')

    def test_help_help(self):
        self.assertHelp('''\
usage: ols-jenkaas... help [-h] [COMMAND [COMMAND ...]]

Describe ols-jenkaas commands.

positional arguments:
  COMMAND     Display help for each command (All if none is given).

optional arguments:
  -h, --help  show this help message and exit
''',
                        ['help'])


class TestApprovedProposalsOptions(unittest.TestCase):

    scenarios = [('bzr', dict(command_cls=commands.ApprovedBzrProposals)),
                 ('git', dict(command_cls=commands.ApprovedGitProposals))]

    def setUp(self):
        super(TestApprovedProposalsOptions, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        approved_cmd = self.command_cls(out=self.out, err=self.err)
        return approved_cmd.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_too_much(self):
        with self.assertRaises(SystemExit):
            self.parse_args(['foo', 'bar'])

    def test_target(self):
        ns = self.parse_args(['bar'])
        self.assertEqual('bar', ns.target)


class TestRunInWorkerOptions(unittest.TestCase):

    def setUp(self):
        super(TestRunInWorkerOptions, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.command = commands.RunInWorker(out=self.out, err=self.err)

    def parse_args(self, args):
        return self.command.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_no_commands(self):
        with self.assertRaises(SystemExit):
            self.parse_args(['container'])

    def test_multiple_commands(self):
        self.parse_args(['container', 'pwd', 'whoami'])
        self.assertEqual('container', self.command.options.container_name)
        self.assertEqual(['pwd', 'whoami'], self.command.options.commands)


class TestRunInWorker(unittest.TestCase):

    def setUp(self):
        super(TestRunInWorker, self).setUp()
        self.lp = fixtures.use_qastaging(self)
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.log_stream = io.StringIO()
        fixtures.override_logging(self, self.log_stream)
        self.command = commands.RunInWorker(out=self.out, err=self.err)
        fixtures.define_uniq_container(self)
        conf = config.VmStack(None)
        # Silence warnings about added hosts by using log level ERROR
        ssh_opts = ('-oLogLevel=ERROR,'
                    '-oUserKnownHostsFile=/dev/null,'
                    '-oStrictHostKeyChecking=no,'
                    '-oIdentityFile={ssh.key}')
        conf.set('ssh.options', ssh_opts)
        conf.store.save()
        conf.store.unload()
        # Pre-build the container or errors in that part are obscure to debug
        self.container = containers.Container(self.container_name)
        self.addCleanup(self.container.teardown)
        self.container.setup()

    def run_command(self, args):
        self.command.parse_args([self.container_name] + args)
        return self.command.run()

    def test_outside_workspace(self):
        os.chdir('..')
        # FIXME: The mounts assume we're below workspace, the end result is
        # broken -- vila 2017-02-06
        ret = self.run_command(['pwd'])
        self.assertEqual(0, ret)
        # '/' below is '/workspace/..' a place where no access is granted
        self.assertEqual('/\n', self.out.getvalue())

    def test_inside_workspace(self):
        os.mkdir('work')
        os.chdir('work')
        ret = self.run_command(['pwd'])
        self.assertEqual(0, ret)
        self.assertEqual('/workspace/work\n', self.out.getvalue())

    def test_below_workspace(self):
        os.makedirs('work/dir/subdir')
        os.chdir('work/dir/subdir')
        ret = self.run_command(['pwd'])
        self.assertEqual(0, ret)
        self.assertEqual('/workspace/work/dir/subdir\n', self.out.getvalue())


class TestLandProposalsOptions(unittest.TestCase):

    scenarios = [('bzr', dict(command_cls=commands.LandApprovedBzrProposal)),
                 ('git', dict(command_cls=commands.LandApprovedGitProposal))]

    def setUp(self):
        super(TestLandProposalsOptions, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.command = self.command_cls(out=self.out, err=self.err)

    def parse_args(self, args):
        return self.command.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_mandatory_args(self):
        self.parse_args(['target', 'work', 'container', 'setup', 'tests'])
        self.assertEqual('target', self.command.options.target)
        self.assertEqual('work', self.command.options.work_dir)
        self.assertEqual('container', self.command.options.container_name)
        self.assertEqual('setup', self.command.options.setup_cmd)
        self.assertEqual('tests', self.command.options.test_cmd)


class TestApprovalCommands(unittest.TestCase):
    """Base class for command related tests.

    This setup the scenarios and the common fixtures.
    """

    scenarios = [('bzr', dict(branch_handler=fixtures.BzrBranchHandler(),
                              proposal_handler=fixtures.BzrProposalHandler(),
                              command_cls=commands.LandApprovedBzrProposal)),
                 ('git', dict(branch_handler=fixtures.GitRepositoryHandler(),
                              proposal_handler=fixtures.GitProposalHandler(),
                              command_cls=commands.LandApprovedGitProposal))]

    def setUp(self):
        super(TestApprovalCommands, self).setUp()
        self.lp = fixtures.use_qastaging(self)
        self.branch_handler.set_identity(self)
        self.branch_handler.set_launchpad_access(self)
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.log_stream = io.StringIO()
        fixtures.override_logging(self, self.log_stream)
        self.command = self.command_cls(out=self.out, err=self.err)

    def run_command(self, args):
        self.command.parse_args(args)
        return self.command.run()


class TestApprovedProposalsErrors(TestApprovalCommands):
    """Test the error scenarios for the *-approved-proposals commands."""

    def setUp(self):
        super(TestApprovedProposalsErrors, self).setUp()
        fixtures.setup_proposal_target(self)
        # Fake running under jenkins so comments on failure can be checked
        self.fake_url = 'http://jenkins.fake/job/x/42'
        os.environ['BUILD_URL'] = self.fake_url

    def assertLogEndsWith(self, expected):
        log = self.log_stream.getvalue()
        self.assertTrue(log.endswith(expected), log)

    def run_command(self, args):
        # For the tests in this class only the target url is needed so fake the
        # other mandatory command args
        dummy_args = ['./work', 'container', 'setup', 'tests']
        self.command.parse_args(args + dummy_args)
        return self.command.run()

    def test_unknown_branch(self):
        unknown_url = self.branch_handler.lp_branch_url('i-dont-exist')
        ret = self.run_command([unknown_url])
        self.assertEqual(1, ret)
        self.assertEqual(self.command.wrong_target.format(unknown_url) + '\n',
                         self.log_stream.getvalue())

    def test_no_proposals(self):
        ret = self.run_command([self.target_url])
        self.assertEqual(0, ret)
        self.assertLogEndsWith('No proposal is approved\n')

    def test_not_approved(self):
        fixtures.create_proposal(self, 'not-approved')
        ret = self.run_command([self.target_url])
        self.assertEqual(0, ret)
        self.assertLogEndsWith('No proposal is approved\n')

    def test_no_commit_message(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True,
                                            commit_msg='')
        ret = self.run_command([self.target_url])
        self.assertEqual(1, ret)
        self.assertLogEndsWith(
            '{} could not be landed: A commit message must be set\n'.format(
                proposal.web_link))
        proposal.lp_refresh()
        self.assertEqual('Needs review', proposal.queue_status)
        last_comment = list(proposal.all_comments)[-1]
        self.assertEqual(
            'A commit message must be set\n{}'.format(self.fake_url),
            last_comment.message_body)

    def test_needs_fixing(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True)
        proposal.createComment(subject='subject2', vote='Needs Fixing',
                               content='there is an issue')
        ret = self.run_command([self.target_url])
        self.assertEqual(1, ret)
        self.assertLogEndsWith(
            '{} could not be landed: Voting criteria not met\n'.format(
                proposal.web_link))
        proposal.lp_refresh()
        self.assertEqual('Needs review', proposal.queue_status)
        last_comment = list(proposal.all_comments)[-1]
        self.assertEqual(
            'Voting criteria not met\n{}'.format(self.fake_url),
            last_comment.message_body)


class TestSetupMergeTarget(TestApprovalCommands):

    def test_existing_dir(self):
        branch = self.branch_handler.factory('target')
        self.addCleanup(branch.delete)
        branch.create()
        branch.update('Something', 'file: foo\nfoo content\n')
        target_url = self.branch_handler.lp_branch_url('target')
        branch.push(target_url)
        os.mkdir('work')
        lp_branch = self.branch_handler.get_branch_from_lp(self.lp, target_url)
        self.addCleanup(lp_branch.lp_delete)
        self.command.setup_merge_target('./work', target_url)
        self.assertTrue(os.path.exists('work/foo'))

    def test_non_existing_subdir(self):
        branch = self.branch_handler.factory('target')
        self.addCleanup(branch.delete)
        branch.create()
        branch.update('Something', 'file: foo\nfoo content\n')
        target_url = self.branch_handler.lp_branch_url('target')
        branch.push(target_url)
        lp_branch = self.branch_handler.get_branch_from_lp(self.lp, target_url)
        self.addCleanup(lp_branch.lp_delete)
        self.command.setup_merge_target('./work/subdir/dir', target_url)
        self.assertTrue(os.path.exists('work/subdir/dir/foo'))

    def test_unknown_branch(self):
        target_url = self.branch_handler.lp_branch_url('i-dont-exist')
        work_dir = './work'
        with self.assertRaises(commands.errors.CommandError):
            self.command.setup_merge_target(work_dir, target_url)
        log = self.log_stream.getvalue()
        # The log provides enough context
        expected = '{} {} failed'.format(target_url, work_dir)
        self.assertTrue(expected in log,
                        '[{}] not found in [{}]'.format(expected, log))


class TestMergeProposedBranch(TestApprovalCommands):

    def setUp(self):
        super(TestMergeProposedBranch, self).setUp()
        fixtures.setup_proposal_target(self)
        # Create the proposal
        self.proposal = fixtures.create_proposal(
            self, 'proposed', approved=True)
        self.work_dir = './work'
        # Prepare the proposal merge
        self.command.setup_merge_target(self.work_dir, self.target_url)

    def test_merge_succeeds(self):
        self.command.merge_proposed_branch(self.work_dir, self.proposal)
        self.assertTrue(os.path.exists('work/proposed'))

    def test_already_merged(self):
        # Until webhooks are used, the current design may trigger an attempt to
        # land the same proposal twice.
        self.command.merge_proposed_branch(self.work_dir, self.proposal)
        self.command.commit_proposal(self.work_dir, self.proposal)
        self.assertFalse(self.command.merge_proposed_branch(
            self.work_dir, self.proposal))

    def test_merge_conflicts(self):
        # Inject a conflicting revision in the work branch
        target = self.branch_handler.factory('work')
        target.update('new content for proposed',
                      'file: proposed\nnew content\n')
        with self.assertRaises(commands.errors.CommandError):
            self.command.merge_proposed_branch(self.work_dir, self.proposal)
        # The log provides enough context
        log = self.log_stream.getvalue()
        expected = 'merge'
        self.assertTrue(expected in log,
                        '[{}] not found in [{}]'.format(expected, log))


class TestCommitProposal(TestApprovalCommands):

    def setUp(self):
        super(TestCommitProposal, self).setUp()
        fixtures.setup_proposal_target(self)
        # Create the proposal
        self.proposal = fixtures.create_proposal(
            self, 'proposed', approved=True,
            # Use a unicode commit message to exercise code paths from local
            # host to launchpad and back as well as vcs commands.
            commit_msg='Unbreakable\xa0Unicode\xa0Space')
        self.work_dir = './work'
        # Merge the proposal
        self.command.setup_merge_target(self.work_dir, self.target_url)
        self.command.merge_proposed_branch(self.work_dir, self.proposal)

    def test_commit_succeeds(self):
        self.command.commit_proposal(self.work_dir, self.proposal)
        # FIXME: Check log ? -- vila 2017-01-24

# MISSINGTEST: What if the commit fails ? Enough feedback ?


class TestPushMergedProposal(TestApprovalCommands):

    def setUp(self):
        super(TestPushMergedProposal, self).setUp()
        fixtures.setup_proposal_target(self)
        # Create the proposal
        self.proposal = fixtures.create_proposal(
            self, 'proposed', approved=True)
        self.work_dir = './work'
        # Merge the proposal and commit
        self.command.setup_merge_target(self.work_dir, self.target_url)
        self.command.merge_proposed_branch(self.work_dir, self.proposal)
        self.command.commit_proposal(self.work_dir, self.proposal)

    def test_push_succeeds(self):
        self.command.push_merged_proposal(self.work_dir, self.target_url)
        # FIXME: Check log ? -- vila 2017-01-24

# MISSINGTEST: What if the push fails ? Enough feedback ?


class TestCommandWithContainer(unittest.TestCase):

    def setUp(self):
        super(TestCommandWithContainer, self).setUp()
        fixtures.isolate_from_disk(self)
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.log_stream = io.StringIO()
        fixtures.override_logging(self, self.log_stream)
        self.command = commands.CommandWithContainer(
            out=self.out, err=self.err)
        fixtures.define_uniq_container(self)
        conf = config.VmStack(None)
        # Silence warnings about added hosts by using log level ERROR
        ssh_opts = ('-oLogLevel=ERROR,'
                    '-oUserKnownHostsFile=/dev/null,'
                    '-oStrictHostKeyChecking=no,'
                    '-oIdentityFile={ssh.key}')
        conf.set('ssh.options', ssh_opts)
        conf.store.save()
        conf.store.unload()
        self.base_container = containers.Container(self.container_name)
        self.addCleanup(self.base_container.teardown)
        ret = self.command.create_worker(self.container_name,
                                         name='worker-' + self.container_name)
        self.addCleanup(self.command.delete_worker)
        self.assertEqual(0, ret)

    def test_command_success(self):
        ret = self.command.run_in_container(['/bin/true'])
        self.assertEqual(0, ret)
        self.assertEqual('', self.out.getvalue())
        self.assertEqual('', self.err.getvalue())

    def test_command_fails(self):
        ret = self.command.run_in_container(['/bin/false'])
        self.assertEqual(1, ret)
        self.assertEqual('', self.out.getvalue())
        self.assertEqual('', self.err.getvalue())

    def test_command_output(self):
        ret = self.command.run_in_container(['whoami'])
        self.assertEqual(0, ret)
        # This got captured by the command.out stream
        self.assertEqual('ubuntu\n', self.out.getvalue())
        self.assertEqual('', self.err.getvalue())

    def test_with_creds(self):
        ret = self.command.run_in_container(['ssh-add', '-l'], with_creds=True)
        self.assertEqual(0, ret)
        # Check that the right key and only that key is provided by the agent.
        ssh_key = self.command.worker.conf.get('ssh.key')
        lines = self.out.getvalue().splitlines()
        assertions.assertLength(self, 1, lines)
        self.assertTrue(ssh_key in lines[0])
        self.assertEqual('', self.err.getvalue())

    def test_without_creds(self):
        ret = self.command.run_in_container(['ssh-add', '-l'],
                                            with_creds=False)
        self.assertEqual(2, ret)
        self.assertEqual('', self.out.getvalue())
        self.assertEqual(
            'Could not open a connection to your authentication agent.\n',
            self.err.getvalue())


class TestSetupVcs(unittest.TestCase):

    def setUp(self):
        super(TestSetupVcs, self).setUp()
        fixtures.isolate_from_disk(self)
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.log_stream = io.StringIO()
        fixtures.override_logging(self, self.log_stream)
        self.command = commands.CommandWithContainer(
            out=self.out, err=self.err)
        fixtures.define_uniq_container(self)
        conf = config.VmStack(None)
        # Silence warnings about added hosts by using log level ERROR
        ssh_opts = ('-oLogLevel=ERROR,'
                    '-oUserKnownHostsFile=/dev/null,'
                    '-oStrictHostKeyChecking=no,'
                    '-oIdentityFile={ssh.key}')
        conf.set('ssh.options', ssh_opts)

    def setup_worker(self, vm_packages):
        conf = config.VmStack(self.container_name)
        conf.set('vm.packages', vm_packages)
        conf.store.save()
        conf.store.unload()
        self.base_container = containers.Container(self.container_name)
        self.addCleanup(self.base_container.teardown)
        ret = self.command.create_worker(self.container_name,
                                         name='worker-' + self.container_name)
        self.addCleanup(self.command.delete_worker)
        self.assertEqual(0, ret)

    def test_config_files_present(self):
        self.setup_worker('bzr')
        ret = self.command.run_in_container(['ls ~/.bazaar/bazaar.conf'])
        self.assertEqual(0, ret)
        # git seems to be installed on lxd images so it's always there
        ret = self.command.run_in_container(['ls', '~/.gitconfig'])
        self.assertEqual(0, ret)

    def test_config_files_absent(self):
        self.setup_worker('')
        ret = self.command.run_in_container(['ls ~/.bazaar/bazaar.conf'])
        self.assertEqual(2, ret)
        # git seems to be installed on lxd images so it's always there
        ret = self.command.run_in_container(['ls', '~/.gitconfig'])
        self.assertEqual(0, ret)


class TestLandings(TestApprovalCommands):

    def setUp(self):
        super(TestLandings, self).setUp()
        fixtures.setup_proposal_target(self)
        fixtures.define_uniq_container(self)
        # Fake running under jenkins so comments on failure can be checked
        self.fake_url = 'http://jenkins.fake/job/x/42'
        os.environ['BUILD_URL'] = self.fake_url

    def assertLogEndsWith(self, expected):
        log = self.log_stream.getvalue()
        self.assertTrue(log.endswith(expected), log)

    def test_landing_succeeds(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True)
        base_container = containers.Container(self.container_name)
        self.addCleanup(base_container.teardown)
        ret = self.run_command([self.target_url, './work', self.container_name,
                                '/bin/true', '/bin/true'])
        self.assertEqual(0, ret)
        self.assertLogEndsWith('{} has landed\n'.format(proposal.web_link))
        proposal.lp_refresh()
        # FIXME: Restore the following when webhooks are used and launchpad
        # trusted to mark the MP as merged (and decorate commits)
        # -- vila 2017-03-30
        self.assertEqual('Merged', proposal.queue_status)

    def test_landing_twice_fails(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True)
        base_container = containers.Container(self.container_name)
        self.addCleanup(base_container.teardown)
        ret = self.run_command([self.target_url, './work', self.container_name,
                                '/bin/true', '/bin/true'])
        self.assertEqual(0, ret)
        self.assertLogEndsWith('{} has landed\n'.format(proposal.web_link))
        # Mark the MP as approved again as it has been marked as Merged.
        proposal.lp_refresh()
        proposal.setStatus(status='Approved', revid=proposal.reviewed_revid)
        ret = self.run_command([self.target_url, './work', self.container_name,
                                '/bin/true', '/bin/true'])
        self.assertEqual(1, ret)
        self.assertLogEndsWith(
            '{} could not be landed: There was nothing to merge\n'.format(
                proposal.web_link))
        proposal.lp_refresh()
        self.assertEqual('Needs review', proposal.queue_status)

    def test_landing_setup_fails(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True)
        base_container = containers.Container(self.container_name)
        self.addCleanup(base_container.teardown)
        ret = self.run_command([self.target_url, './work', self.container_name,
                                '/bin/false', '/bin/true'])
        self.assertEqual(1, ret)
        self.assertLogEndsWith(
            '{} could not be landed: Project setup failed\n'.format(
                proposal.web_link))
        proposal.lp_refresh()
        self.assertEqual('Needs review', proposal.queue_status)
        last_comment = list(proposal.all_comments)[-1]
        self.assertEqual('Project setup failed\n{}'.format(self.fake_url),
                         last_comment.message_body)

    def test_landing_tests_fails(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True)
        base_container = containers.Container(self.container_name)
        self.addCleanup(base_container.teardown)
        ret = self.run_command([self.target_url, './work', self.container_name,
                                '/bin/true', '/bin/false'])
        self.assertEqual(1, ret)
        self.assertLogEndsWith(
            '{} could not be landed: Running landing tests failed\n'.format(
                proposal.web_link))
        proposal.lp_refresh()
        self.assertEqual('Needs review', proposal.queue_status)
        last_comment = list(proposal.all_comments)[-1]
        self.assertEqual(
            'Running landing tests failed\n{}'.format(
                self.fake_url),
            last_comment.message_body)

    def test_landing_conflicts_fails(self):
        proposal = fixtures.create_proposal(self, 'approved', approved=True)
        # Make a conflicting change in the target
        target = self.branch_handler.factory('target')
        target.update('new content for approved',
                      'file: approved\nnew content\n')
        target.push(self.target_url)
        ret = self.run_command([self.target_url, './work', self.container_name,
                                '/bin/true', '/bin/true'])
        self.assertEqual(1, ret)
        self.assertLogEndsWith(
            '{} could not be landed: Merging failed\n'.format(
                proposal.web_link))
        proposal.lp_refresh()
        self.assertEqual('Needs review', proposal.queue_status)
        last_comment = list(proposal.all_comments)[-1]
        self.assertEqual('Merging failed\n{}'.format(self.fake_url),
                         last_comment.message_body)
